package djmil.cordacheckers.gameproposal;

import java.util.Arrays;
import java.util.List;
import java.util.Map;

import djmil.cordacheckers.contracts.GameProposalContract;
import djmil.cordacheckers.states.GameProposalResolutionState;
import djmil.cordacheckers.states.GameProposalState;
import djmil.cordacheckers.states.GameProposalResolutionState.Resolution;
import net.corda.v5.application.flows.ClientRequestBody;
import net.corda.v5.application.flows.ClientStartableFlow;
import net.corda.v5.application.flows.CordaInject;
import net.corda.v5.application.flows.InitiatingFlow;
import net.corda.v5.application.marshalling.JsonMarshallingService;
import net.corda.v5.application.membership.MemberLookup;
import net.corda.v5.application.messaging.FlowMessaging;
import net.corda.v5.application.messaging.FlowSession;
import net.corda.v5.base.annotations.Suspendable;
import net.corda.v5.base.exceptions.CordaRuntimeException;
import net.corda.v5.ledger.utxo.Command;
import net.corda.v5.ledger.utxo.StateAndRef;
import net.corda.v5.ledger.utxo.TransactionState;
import net.corda.v5.ledger.utxo.UtxoLedgerService;
import net.corda.v5.ledger.utxo.transaction.UtxoSignedTransaction;
import net.corda.v5.ledger.utxo.transaction.UtxoTransactionBuilder;

import static java.util.Map.entry;
import static java.util.stream.Collectors.toList;

import java.time.Duration;
import java.time.Instant;

@InitiatingFlow(protocol = "game-proposal-action")
public class ActionFlow implements ClientStartableFlow {

    @CordaInject
    public JsonMarshallingService jsonMarshallingService;

    @CordaInject
    public UtxoLedgerService ledgerService;

    @CordaInject
    public FlowMessaging flowMessaging;

    @CordaInject
    public MemberLookup memberLookup;

    private final static Map<GameProposalResolutionState.Resolution, Command> resoultion2command = Map.ofEntries(
        entry(GameProposalResolutionState.Resolution.CANCEL, new GameProposalContract.Cancel()),
        entry(GameProposalResolutionState.Resolution.REJECT, new GameProposalContract.Reject()),
        entry(GameProposalResolutionState.Resolution.ACCEPT, new GameProposalContract.Accept())
    );

    @Override
    @Suspendable
    public String call(ClientRequestBody requestBody) {
        ActionFlowArgs args = requestBody.getRequestBodyAs(jsonMarshallingService, ActionFlowArgs.class);

        /*
         * Look up the latest unconsumed ChatState with the given id.
         * Note, this code brings all unconsumed states back, then filters them. This is an
         * inefficient way to perform this operation when there are a large number of chats
         */
        List<StateAndRef<GameProposalState>> stateAndRefs = this.ledgerService
            .findUnconsumedStatesByType(GameProposalState.class);

        List<StateAndRef<GameProposalState>> stateAndRefsWithId = stateAndRefs
            .stream()
            .filter(sar -> sar.getState().getContractState().getId().equals(args.gameProposalUuid))
            .collect(toList());
        if (stateAndRefsWithId.size() != 1) 
            throw new CordaRuntimeException("Multiple or zero GameProposal states with id " + args.gameProposalUuid + " found");
        
        StateAndRef<GameProposalState> stateAndRef = stateAndRefsWithId.get(0);
        TransactionState<GameProposalState> trxState = stateAndRef.getState();
        GameProposalState state = trxState.getContractState();

        GameProposalResolutionState outputState = new GameProposalResolutionState(
            GameProposalResolutionState.Resolution.valueOf(args.action),
            state.getParticipants()
            );

        /*
         * Build draft trx
         */
        UtxoTransactionBuilder txBuilder = ledgerService.createTransactionBuilder()
            .setNotary(trxState.getNotaryName())
            .setTimeWindowBetween(Instant.now(), Instant.now().plusMillis(Duration.ofDays(1).toMillis()))
            .addInputState(stateAndRef.getRef())
            .addOutputState(outputState)
            .addCommand(resoultion2command.get(outputState.outcome))
            .addSignatories(state.getParticipants());

        UtxoSignedTransaction signedTransaction = txBuilder.toSignedTransaction();

        FlowSession session = flowMessaging.initiateFlow(
            outputState.outcome == Resolution.CANCEL ? state.getRecipient() : state.getSender() // TODO: readability
            );

        List<FlowSession> sessionsList = Arrays.asList(session);

        ledgerService.finalize(signedTransaction, sessionsList);

        return args.action+"ED"; // REJECT+ED
    }
    
}
